/*
******************************************************************************

   LibBLT: plain functions

   Copyright (C) 2001 Christoph Egger   [Christoph_Egger@t-online.de]
   Copyright (C) 2001 Brian S. Julin    [bri@tull.umassp.edu]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <ggi/internal/blt.h>
#include <ggi/internal/blt_debug.h>


#if 0
	/* useful code fragment for debugging */
	do {
	  char *foo;
	  ggiGAanprintf(vis, reqlist, 8191, "", &foo);
	  fprintf(stderr, "reqlist before check:\n%s\n", foo);
	  free(foo);
	} while(0);
#endif


/* LibGAlloc advanced interface.
 */
int ggiBltAdd(ggiGA_resource_list *reqlist,
              struct ggiGA_resource_props **props, /* motor, bobs, carbs */
              enum ggiGA_resource_type *types, int numresources,
              ggiGA_resource_handle **handles) {
	int i, tag, have_primary_carb;

#warning TODO error checking

	if (handles == NULL) return GGI_EARGINVAL;
	if (*handles == NULL) 
		*handles = calloc(numresources, sizeof(**handles));

	/* Add motors */
	tag = 0;
	for (i = 0; i < numresources; i++) {
		if (types[i] & GA_RT_MOTOR) {
			int rc;
			rc = ggiGAAdd(reqlist, props[i], types[i], handles[i]);
			if (!tag) 
				tag = ggiGASetAutoTag(*reqlist, *(handles[i]));
			else 
				ggiGASetTag(*(handles[i]), tag);
		}
	}
	have_primary_carb = 0;
	for (i = 0; i < numresources; i++) {
		if (types[i] & GA_RT_MOTOR) continue;
		if (!(types[i] & GA_RT_CARB)) {
			ggiGAAdd(reqlist, props[i], types[i], handles[i]);
			continue;
		}
		if (!have_primary_carb) {
			ggiGAAdd(reqlist, props[i], types[i], handles[i]);
			have_primary_carb = 1;
			ggiGASetTag(*(handles[i]), tag);
			continue;

		}
		ggiGAAddSeeabove(reqlist, props[i], types[i], handles[i]);
		ggiGASetTag(*(handles[i]), tag);
	}
	return 0;
}

ggiBlt_t ggiBltCreate(ggi_visual_t vis, enum ggiGA_storage_type stype,
		      int numrows, enum ggi_ll_rop rop, enum ggiBlt_flags flags)
{
	struct ggiGA_resource_props props_m, props_c, *props_p[2];
	enum ggiGA_resource_type rt[2];
	ggiGA_resource_handle hand_m, hand_c, *hand_p[2];
	ggiGA_resource_list reqlist, reslist;
	ggiBlt_t res;
	int rc;
	reqlist = reslist = NULL;

	/* Prime the pump -- in case no mode resource in visual yet. */
	if (*(LIBGGI_BLTEXT(vis)->reqlist) == NULL) {
		BLTDPRINT_CORE("Jump starting LibGAlloc\n");
		ggiGAGet(vis, LIBGGI_BLTEXT(vis)->reqlist);
	}       /* if */

	/* Copy the current resource list into the local reqlist,
	 * and add our motor to it, save the handle.
	 */
	rc = ggiGAGet(vis, &reqlist);
	LIB_ASSERT(rc == GALLOC_OK);
	if (rc != GALLOC_OK) return NULL;

	/* Construct props for our motor and carb */
	props_p[0] = &props_m; 
	props_p[1] = &props_c;
	hand_p[0] = &hand_m;
	hand_p[1] = &hand_c;
	rt[0] = GA_RT_MOTOR | GA_RT_BOB_DONTCARE;
	rt[1] = GA_RT_CARB | GA_RT_BOB_DONTCARE;
	ggiGAClearMotorProperties(&props_m);
	ggiGAClearCarbProperties(&props_c);
	props_c.storage_need = props_c.storage_ok = stype;
	props_c.sub.carb.must.rop = props_c.sub.carb.can.rop = rop;

	ggiBltAdd(&reqlist, props_p, rt, 2, hand_p);

        BLTDPRINT_CORE("Checking global reqlist\n");
        rc = ggiGACheck(vis, reqlist, NULL);
        BLTDPRINT("%s:%s:%i: return code of ggiGACheck(): %i\n", 
		  DEBUG_INFO, rc);
        if (rc != GALLOC_OK) goto err0;

        rc = ggiGASet(vis, reqlist, &reslist);
        BLTDPRINT("%s:%s:%i: return code of ggiGASet(): %i\n", DEBUG_INFO, rc);
        if (rc != GALLOC_OK) goto err1;

	/* Create the blt object -- pass the carb, tie will find the motor. */
	*(hand_p[0]) = ggiGAHandle(reslist, reqlist, *(hand_p[0]));
	*(hand_p[1]) = ggiGAHandle(reslist, reqlist, *(hand_p[1]));
	res = ggiBltTie(*hand_p, 2);
	if (res == NULL) goto err2;

        rc = ggiGAEmptyList(&reqlist);
        LIB_ASSERT(rc == GALLOC_OK);

        /* perform touch-up surgery on the reqlist */
        rc = ggiGASpliceList(LIBGGI_BLTEXT(vis)->reqlist, &reslist);
        LIB_ASSERT(rc == GALLOC_OK);

        BLTDPRINT("%s:%s:%i: completed successfully\n", DEBUG_INFO);
        return res;
 err2:
#warning TODO after LibGAlloc rewrite
 err1:
        BLTDPRINT("%s:%s:%i: err1\n", DEBUG_INFO);
        rc = ggiGAEmptyList(&reslist);
        LIB_ASSERT(rc == GALLOC_OK);
 err0:
        BLTDPRINT("%s:%s:%i: err0\n", DEBUG_INFO);
        rc = ggiGAEmptyList(&reqlist);
        LIB_ASSERT(rc == GALLOC_OK);

        BLTDPRINT("%s:%s:%i: returning with error %i\n", DEBUG_INFO, rc);
        return NULL;
}

int ggiBltDestroy(ggiBlt_t *blt) {
#warning TODO after LibGAlloc rewrite
	free(*blt);
	*blt = NULL;
	return 0;
}

ggiBlt_source_t ggiBltCreateBob(ggi_visual_t vis, ggiBlt_t forthis,
				enum ggiGA_storage_type stype, int w, int h,
				int qty, ggi_graphtype gt) {
	ggiBlt_source_t res;
	ggiBlt_bob_t bob;
	struct ggiGA_resource_props props;
	ggiGA_resource_list reslist, reqlist;
	ggiGA_resource_handle hand_t, hand_c;
	int rc;

	if (vis == NULL) return NULL;
	if (forthis == NULL) return NULL;
	if (vis != forthis->vis) return NULL;
	if ((w <= 0) || (h <= 0)) return NULL; 
	reslist = reqlist = NULL;

#warning TODO: more error checking

	ggiGAGet(vis, &reqlist);
	ggiGAClearTankProperties(&props);
	props.qty = qty;
	props.storage_need = props.storage_ok = stype;
	props.size.area.x = w;
	props.size.area.y = h;
	props.sub.tank.gt = gt;
	ggiGAAdd(&reqlist, &props, GA_RT_BOB_DONTCARE, &hand_t);
#warning TODO: use a mutual.
	ggiGAAddSeeabove(&reqlist, ggiGAGetProperties(forthis->def_carb_h), 
			 GA_RT_BOB_DONTCARE, &hand_c);

        BLTDPRINT_CORE("Checking global reqlist\n");
        rc = ggiGACheck(vis, reqlist, NULL);
        BLTDPRINT("%s:%s:%i: return code of ggiGACheck(): %i\n",
                  DEBUG_INFO, rc);
        if (rc != GALLOC_OK) goto err0;

        rc = ggiGASet(vis, reqlist, &reslist);
        BLTDPRINT("%s:%s:%i: return code of ggiGASet(): %i\n", DEBUG_INFO, rc);
        if (rc != GALLOC_OK) goto err1;

	hand_t = ggiGAHandle(reslist, reqlist, hand_t);
	hand_c = ggiGAHandle(reslist, reqlist, hand_c);

	res = calloc(1, sizeof(struct ggiBlt_source));
	if (res == NULL) goto err1;
	bob = res->src1 = calloc(1, sizeof(struct ggiBlt_bob));
	if (res == NULL) goto err2;

	bob->tank_h = hand_t;
	bob->carb_h = hand_c;
	ggiBltSetSourceIdx(res, 0);

        rc = ggiGAEmptyList(&reqlist);
        LIB_ASSERT(rc == GALLOC_OK);

        /* perform touch-up surgery on the reqlist */
        rc = ggiGASpliceList(LIBGGI_BLTEXT(vis)->reqlist, &reslist);
        LIB_ASSERT(rc == GALLOC_OK);

	return (res);
	
 err2:  
	/* Release the resources */
#warning TODO after LibGAlloc rewrite
	free(res);
 err1:
        BLTDPRINT("%s:%s:%i: err1\n", DEBUG_INFO);
        rc = ggiGAEmptyList(&reslist);
        LIB_ASSERT(rc == GALLOC_OK);
 err0:
        BLTDPRINT("%s:%s:%i: err0\n", DEBUG_INFO);
        rc = ggiGAEmptyList(&reqlist);
        LIB_ASSERT(rc == GALLOC_OK);

        BLTDPRINT("%s:%s:%i: returning with error %i\n", DEBUG_INFO, rc);
        return NULL;
}

int ggiReuseBob(ggiBlt_source_t bob, ggiBlt_t forthis) {
#warning TODO after LibGAlloc rewrite
	return -1;
}

int ggiBltDestroyBob(ggiBlt_source_t *bob) {
	/* Release the resources */
#warning TODO after LibGAlloc rewrite
	free((*bob)->src1);
	free(*bob);
	bob = NULL;
	return 0;
}

/* Functions for BOB based ggiBlt_source objects */
int ggiBltSetBobRop(ggiBlt_source_t bob, enum ggi_ll_rop rop) {
#warning TODO error checking
	((ggiBlt_bob_t)(bob->src1))->rop = rop;
	return 0;
}
int ggiBltGetBobRop(ggiBlt_source_t bob, enum ggi_ll_rop *rop) {
#warning TODO error checking
	*rop = ((ggiBlt_bob_t)(bob->src1))->rop;
	return 0;
}
int ggiBltSetBobKeyValue(ggiBlt_source_t bob, ggi_pixel keyval) {
#warning TODO error checking
	((ggiBlt_bob_t)(bob->src1))->keyval = keyval;
	return 0;
}
int ggiBltGetBobKeyValue(ggiBlt_source_t bob, ggi_pixel *keyval) {
#warning TODO error checking
	*keyval = ((ggiBlt_bob_t)(bob->src1))->keyval;
	return 0;
}
int ggiBltSetBobKeyMask(ggiBlt_source_t bob, ggi_pixel keymask) {
#warning TODO error checking
	((ggiBlt_bob_t)(bob->src1))->keymask = keymask;
	return 0;
}
int ggiBltGetBobKeyMask(ggiBlt_source_t bob, ggi_pixel *keymask) {
#warning TODO error checking
	*keymask = ((ggiBlt_bob_t)(bob->src1))->keymask;
	return 0;
}
int ggiBltSetBobFlags(ggiBlt_source_t bob, enum ggiBlt_bob_flags flags) {
#warning TODO error checking
	((ggiBlt_bob_t)(bob->src1))->flags = flags;
	return 0;
}
int ggiBltGetBobFlags(ggiBlt_source_t bob, enum ggiBlt_bob_flags *flags) {
#warning TODO error checking
	*flags = ((ggiBlt_bob_t)(bob->src1))->flags;
	return 0;
}

/* Functions that may work on buf/visual/other types of sources as well */
int ggiBltSetSourceIdx(ggiBlt_source_t bob, int idx) {
#warning this code should handle all types of sources, not just bobs
  ggiBlt_bob_t realbob;

  realbob = (ggiBlt_bob_t)bob->src1;
  /* This needs to support idx != 0 also, of course :-) */
  realbob->read =  ggiGAGetProperties(realbob->tank_h)->sub.tank.read;
  realbob->write = ggiGAGetProperties(realbob->tank_h)->sub.tank.write;
  return 0;
}
int ggiBltGetSourceIdx(ggiBlt_source_t bob, int idx) {
#warning TODO
  return -1;
}

/* Blt interface */
int ggiBltSetIdx(ggiBlt_t blt, int idx) {
	if (blt->trans == NULL) return GGI_EARGINVAL;
	bo_set_in_idx(blt->trans, idx);
	return 0;
}
int ggiBltGetIdx(ggiBlt_t blt, int *idx) {
	if (blt->trans == NULL) return GGI_EARGINVAL;
//	*idx = bo_get_in_idx(blt->getidx, idx);  
	return 0;
}
int ggiBltGo(ggiBlt_t blt, int start, int numrows) {
	if (blt->sink == NULL) return GGI_EARGINVAL;
	bo_set_out_idx(blt->trans, start);
	return bo_go(blt->trans, blt->sink, numrows);  
}
int ggiBltAt(ggiBlt_t blt) {
	if (blt->sink == NULL) return GGI_EARGINVAL;  
//	return bo_at(blt->sink);  
	return 0;
}

int ggiBltPutBatchop(ggiBlt_t into, batchop_t from, int numrows) {
#warning TODO
	return -1;
}

int ggiBltPutBlit(ggiBlt_t blt, ggiBlt_source_t source,
		  long x, long y, int w, int h) {
	u_int32_t opcode;
	struct batchparm bp[6] = {
	  {
	    parmtype:	BLT_PT_DSTX,
	    block:	&x,
	    datasize:	sizeof(int) * 8,
	    access:	0,
	    offset:	0,
	    transient:	0,
	    step:	0
	  },
	  {
	    parmtype:	BLT_PT_DSTY,
	    block:	&y,
	    datasize:	sizeof(int) * 8,
	    access:	0,
	    offset:	0,
	    transient:	0,
	    step:	0
	  },
	  {
	    parmtype:	BLT_PT_SRCW,
	    block:	&w,
	    datasize:	sizeof(int) * 8,
	    access:	0,
	    offset:	0,
	    transient:	0,
	    step:	0
	  },
	  {
	    parmtype:	BLT_PT_SRCH,
	    block:	&h,
	    datasize:	sizeof(int) * 8,
	    access:	0,
	    offset:	0,
	    transient:	0,
	    step:	0
	  },
	  {
	    parmtype:	GGI_PT_IS_OPCODE,
	    block:	&opcode,
	    datasize:	32,
	    access:	0,
	    offset:	0,
	    transient:	0,
	    step:	0
	  },
	  {
	    parmtype:	BLT_PT_SRCADDR,
	    block:	&(((ggiBlt_bob_t)(source->src1))->read),
	    datasize:	sizeof(ggi_pixel) * 8,
	    access:	0,
	    offset:	0,
	    transient:	0,
	    step:	0
	  }
	};
	struct batchop bo = {
	  parms: bp,
	  numparms: 6,
	};

	opcode = BLT_OP_SIMPLEBLT;
	bo_set_out_idx(((batchop_t)&bo), 0);

	bo_match(((batchop_t)&bo), blt->trans, NULL);
	bo_go(((batchop_t)&bo), blt->trans, 1);
	return -1;
}

int ggiBltPutReverseBlit(ggiBlt_t blt, ggiBlt_source_t dest,
                         long x, long y, int w, int h)
{
#warning TODO
	return -1;
}

int ggiBltPutStretchBlit(ggiBlt_t blt, ggiBlt_source_t source,
			 long x, long y, int sw, int sh, long dw, long dh) {
#warning TODO
	return -1;
}

int ggiBltPutZBlit(ggiBlt_t blt, ggiBlt_source_t source,
		   long x, long y, int w, int h, int ztl, int zbr) {
#warning TODO
	return -1;
}

int ggiBltPutABlit(ggiBlt_t blt, ggiBlt_source_t source,
		   long x, long y, int w, int h, int atl, int abr) {
#warning TODO
	return -1;
}

int ggiBltAlterBobIdx(ggiBlt_t blt, int *bobidx, int numrows) {
#warning TODO
	return -1;
}
int ggiBltAlterCoord(ggiBlt_t blt, long *x, long *y, int numrows) {
#warning TODO
	return -1;
}
int ggiBltAlterBox(ggiBlt_t blt, long *x, long *y, int *w, int *h,
                   int numrows) {
#warning TODO
	return -1;
}
int ggiBltAlterStretchBox(ggiBlt_t blt, long *x, long *y, int *w, int *h,
                          int numrows) {
#warning TODO
	return -1;
}
int ggiBltAlterZ(ggiBlt_t blt, int *ztl, int *zbr, int numrows) {
#warning TODO
	return -1;
}
int ggiBltAlterA(ggiBlt_t blt, int *atl, int *abr, int numrows) {
#warning TODO
	return -1;
}


/* Set the default base for the coordinates: Either pixels or dots
 */
int ggiBltSetCoordBase(ggiBlt_t blt, enum ggi_ll_coordbase cb) {
#warning TODO
	return 0;
}
enum ggi_ll_coordbase ggiBltGetCoordBase(ggiBlt_t blt) {
#warning TODO
	return 0;
}

